from pathlib import Path
import time
import click

from stochasticx.deployment.deployments import (
    Deployment,
    DeploymentsClient,
    Instances
)

from stochasticx.instances.instance_types import InstanceType
from stochasticx.models.models import Models
from stochasticx.utils.parse_utils import print_table
from stochasticx.utils.preferences import Preferences, AppModes
from stochasticx.utils.docker import (
    start_container,
    exists_running_container,
    stop_and_remove_container,
    get_logs_container,
    get_open_ports_container,
    exists_container,
    auto_port_selection,
    wait_container_is_up,
)
from stochasticx.utils.gpu_utils import is_nvidia_gpu_available
from stochasticx.deployment.deployments import (
    LocalDeploymentsClient,
    LocalSequenceClassificationDeployment,
    LocalQuestionAnsweringDeployment,
    LocalTokenClassificationDeployment,
    LocalTranslationDeployment,
    LocalSummarizationDeployment,
    LocalTextGenerationDeployment,
)
from stochasticx.constants.docker_images import DockerImages, ContainerNames
from stochasticx.utils.auth_utils import AuthUtils
import sys

from stochasticx.utils.stat_controller import EventLogger


@click.group(name="deployments")
@click.option(
    "--port",
    default=9000,
    help="Port which Stochastic local inference container will run on",
)
def deployments(port):
    try:
        AuthUtils.get_auth_headers()
    except:
        sys.exit()

    preferences = Preferences.load()

    if (
        preferences.current_mode == AppModes.LOCAL and
        not exists_running_container(ContainerNames.INFERENCE)
    ):
        click.secho(
            "\n[+] We are configuring the environment for inference. If it is the first time, it will take some minutes\n",
            fg="yellow",
            bold=True,
        )

        original_port = port
        port, is_replace = auto_port_selection(port, ContainerNames.INFERENCE)
        if is_replace:
            click.secho(
                f"[+] Port {original_port} is using on another service, port {port} will be selected to run stochasticx model inference",
                fg="blue",
                bold=True,
            )

        port2, _ = auto_port_selection(
            5051, ContainerNames.INFERENCE, selected_ports=[port]
        )
        port3, _ = auto_port_selection(
            9001, ContainerNames.INFERENCE, selected_ports=[port, port2]
        )
        port4, _ = auto_port_selection(
            9002, ContainerNames.INFERENCE, selected_ports=[port, port2, port3]
        )

        start_container(
            docker_image=DockerImages.INFERENCE,
            ports={"5000": port2, "8000": port, "8001": port3, "8002": port4},
            container_name=ContainerNames.INFERENCE,
            detach=True,
            gpu=is_nvidia_gpu_available(),
            network_links={"x-local": "x-local"},
            volumes_from=["x-local"],
        )

        click.echo("[+] Setting up preferences...")
        preferences = Preferences.load()
        preferences.current_mode = AppModes.LOCAL
        preferences.local_inference_url = f"http://127.0.0.1:{port}"
        preferences.local_deployment_url = f"http://127.0.0.1:{port2}"
        Preferences.save(preferences)

        local_url = f"{preferences.local_deployment_url}/model"

        wait_container_is_up(local_url)

        click.secho(
            f"[+] x-local-inference server running in the port {port}",
            fg="green",
            bold=True,
        )
        EventLogger.log_event("deployments_local")


@click.command(name="ls")
def ls_deployments():
    preferences = Preferences.load()

    click.secho("\n[+] Collecting all deployments\n", fg="blue", bold=True)
    if preferences.current_mode == AppModes.LOCAL:
        columns, values = LocalDeploymentsClient.get_deployments()
        EventLogger.log_event("deployments_local_ls")
    else:
        columns, values = DeploymentsClient.get_deployments(fmt="table")
        EventLogger.log_event("deployments_cloud_ls")
    print_table(columns, values)
    
    click.secho("\n[+] Client URL or API key will be available once the deployment is running. Execute the following command to get them: stochasticx inference inspect --deployment_name deployment_name\n", fg="yellow", bold=True)


@click.group(name="deploy")
def deploy():
    pass


@click.command(name="launch") 
@click.option('--name', required=True, help='Deployment name')
@click.option('--instance_type', required=True, help='Instance type')
@click.option('--model_name', required=True, default="hf", show_default=True, help='Model ID')
@click.option('--model_type', required=False, help='Model type')
@click.option('--sub_type', required=False, help='Task or submodel type')
@click.option('--inference_file_path', required=False, help='Inference file path')
@click.option('--docker_image', required=False, help='Custom Docker image')
@click.option('--docker_registry_username', required=False, help='Docker registry username to download the Docker image')
@click.option('--docker_registry_password', required=False, help='Docker registry password to download the Docker image')
def launch_deployment(name, instance_type, model_name, model_type, sub_type, inference_file_path, docker_image, docker_registry_username, docker_registry_password):
    preferences = Preferences.load()
    click.secho("\n[+] Creating deployment...", fg='blue', bold=True)
    
    if preferences.current_mode == AppModes.LOCAL:
        click.secho("\n[+] Command not available in the local version", fg='red', bold=True)
        EventLogger.log_event("deployments_local_launch")
    else:
        valid_instances = [InstanceType.G4DN_XLARGE]
        assert instance_type in valid_instances, "Your instance type is not valid. Valid instances: {}".format(valid_instances)
        
        DeploymentsClient.create_deployment(
            name=name,
            instance_type=instance_type,
            model_name=model_name,
            model_type=model_type,
            task_type=sub_type,
            inference_file_path=inference_file_path,
            docker_image=docker_image,
            docker_registry_username=docker_registry_username,
            docker_registry_password=docker_registry_password
        )
        EventLogger.log_event("deployments_cloud_launch")

    click.secho("\n[+] Deployed", fg='green', bold=True)

@click.command(name="delete") 
@click.option('--name', required=True, help='Deployment ID')
@click.option('--model_name_or_id', required=False, help='Model ID')
@click.option('--model_type', required=False, default="hf", show_default=True, help='Model ID')
def delete_deployment(name, model_name_or_id, model_type):
    preferences = Preferences.load()

    click.secho("\n[+] Deleting deployment...", fg="blue", bold=True)

    if preferences.current_mode == AppModes.LOCAL:
        LocalDeploymentsClient.delete_deployment(
            model_type=model_type, model_name=model_name_or_id
        )
        EventLogger.log_event("deployments_local_delete")
    else:
        DeploymentsClient.delete_deployment(name)
        EventLogger.log_event("deployments_cloud_delete")

    click.secho("[+] Deleted\n", fg="green", bold=True)


@click.command(name="sequence_classification")
@click.option("--model_name_or_id", required=True, help="Model ID")
@click.option(
    "--model_type", required=False, default="hf", show_default=True, help="Model ID"
)
@click.option("--instance_type", required=False, default=None, help="Instance type")
@click.option(
    "--max_batch_size",
    required=False,
    default=8,
    show_default=True,
    help="Maximum batch size",
)
@click.option(
    "--max_seq_length",
    required=False,
    default=128,
    show_default=True,
    help="Maximum source length",
)
def sequence_classification(
    model_name_or_id, model_type, instance_type, max_batch_size, max_seq_length
):
    preferences = Preferences.load()

    click.secho("\n[+] Starting deployment...", fg="blue", bold=True)

    if preferences.current_mode == AppModes.LOCAL:
        LocalSequenceClassificationDeployment(
            model_name=model_name_or_id,
            type=model_type,
            cuda=is_nvidia_gpu_available(),
            user_params={
                "max_batch_size": max_batch_size,
                "max_seq_length": max_seq_length,
            },
        ).start_deployment()
        EventLogger.log_event("deployments_local_sequence_classification")
    else:
        click.secho("\n[+] Command not available in the cloud version", fg='red', bold=True)
        EventLogger.log_event("deployments_cloud_sequence_classification")

    click.secho("[+] Model deployed\n", fg="green", bold=True)


@click.command(name="question_answering")
@click.option("--model_name_or_id", required=True, help="Model ID")
@click.option(
    "--model_type", required=False, default="hf", show_default=True, help="Model ID"
)
@click.option("--instance_type", required=False, default=None, help="Instance type")
@click.option(
    "--max_batch_size", default=8, show_default=True, help="Maximum batch size"
)
@click.option(
    "--max_seq_length", default=128, show_default=True, help="Maximum sequence length"
)
def question_answering(
    model_name_or_id, model_type, instance_type, max_batch_size, max_seq_length
):
    preferences = Preferences.load()

    click.secho("\n[+] Starting deployment...", fg="blue", bold=True)

    if preferences.current_mode == AppModes.LOCAL:
        LocalQuestionAnsweringDeployment(
            model_name=model_name_or_id,
            type=model_type,
            cuda=is_nvidia_gpu_available(),
            user_params={
                "max_batch_size": max_batch_size,
                "max_seq_length": max_seq_length,
            },
        ).start_deployment()
        EventLogger.log_event("deployments_local_question_answering")
    else:
        click.secho("\n[+] Command not available in the cloud version", fg='red', bold=True)
        EventLogger.log_event("deployments_cloud_question_answering")

    click.secho("[+] Model deployed\n", fg="green", bold=True)


@click.command(name="summarization")
@click.option("--model_name_or_id", required=True, help="Model ID")
@click.option(
    "--model_type", required=False, default="hf", show_default=True, help="Model ID"
)
@click.option("--instance_type", required=False, default=None, help="Instance type")
@click.option(
    "--max_batch_size", default=8, show_default=True, help="Maximum batch size"
)
@click.option(
    "--max_source_length", default=128, show_default=True, help="Maximum source length"
)
@click.option("--source_prefix", default="", show_default=True, help="Source prefix")
@click.option("--lang", default="en", show_default=True, help="Language")
def summarization(
    model_name_or_id,
    model_type,
    instance_type,
    max_batch_size,
    max_source_length,
    source_prefix,
    lang,
):
    preferences = Preferences.load()

    click.secho("\n[+] Starting deployment...", fg="blue", bold=True)

    if preferences.current_mode == AppModes.LOCAL:
        LocalSummarizationDeployment(
            model_name=model_name_or_id,
            type=model_type,
            cuda=is_nvidia_gpu_available(),
            user_params={
                "max_batch_size": max_batch_size,
                "max_source_length": max_source_length,
                "lang": lang,
                "source_prefix": source_prefix,
            },
        ).start_deployment()
        EventLogger.log_event("deployments_local_summarization")
    else:
        click.secho("\n[+] Command not available in the cloud version", fg='red', bold=True)
        EventLogger.log_event("deployments_cloud_summarization")

    click.secho("[+] Model deployed\n", fg="green", bold=True)


@click.command(name="translation")
@click.option("--model_name_or_id", required=True, help="Model ID")
@click.option(
    "--model_type", required=False, default="hf", show_default=True, help="Model ID"
)
@click.option("--instance_type", required=False, default=None, help="Instance type")
@click.option(
    "--max_batch_size", default=8, show_default=True, help="Maximum batch size"
)
@click.option(
    "--max_source_length", default=128, show_default=True, help="Maximum source length"
)
@click.option("--source_prefix", default="", show_default=True, help="Source prefix")
@click.option("--src_lang", default="en", show_default=True, help="Source language")
@click.option("--tgt_lang", default="de", show_default=True, help="Target language")
def translation(
    model_name_or_id,
    model_type,
    instance_type,
    max_batch_size,
    max_source_length,
    source_prefix,
    src_lang,
    tgt_lang,
):
    preferences = Preferences.load()

    click.secho("\n[+] Starting deployment...", fg="blue", bold=True)

    if preferences.current_mode == AppModes.LOCAL:
        LocalTranslationDeployment(
            model_name=model_name_or_id,
            type=model_type,
            cuda=is_nvidia_gpu_available(),
            user_params={
                "max_batch_size": max_batch_size,
                "max_source_length": max_source_length,
                "src_lang": src_lang,
                "tgt_lang": tgt_lang,
                "source_prefix": source_prefix,
            },
        ).start_deployment()
        EventLogger.log_event("deployments_local_translation")
    else:
        click.secho("\n[+] Command not available in the cloud version", fg='red', bold=True)
        EventLogger.log_event("deployments_cloud_translation")

    click.secho("[+] Model deployed\n", fg="green", bold=True)


@click.command(name="token_classification")
@click.option("--model_name_or_id", required=True, help="Model ID")
@click.option(
    "--model_type", required=False, default="hf", show_default=True, help="Model ID"
)
@click.option("--instance_type", required=False, default=None, help="Instance type")
@click.option(
    "--max_batch_size", default=8, show_default=True, help="Maximum batch size"
)
@click.option(
    "--max_seq_length", default=128, show_default=True, help="Maximum source length"
)
def token_classification(
    model_name_or_id, model_type, instance_type, max_batch_size, max_seq_length
):
    preferences = Preferences.load()

    click.secho("\n[+] Starting deployment...", fg="blue", bold=True)

    if preferences.current_mode == AppModes.LOCAL:
        LocalTokenClassificationDeployment(
            model_name=model_name_or_id,
            type=model_type,
            cuda=is_nvidia_gpu_available(),
            user_params={
                "max_batch_size": max_batch_size,
                "max_seq_length": max_seq_length,
            },
        ).start_deployment()
        EventLogger.log_event("deployments_local_token_classification")
    else: 
        click.secho("\n[+] Command not available in the cloud version", fg='red', bold=True)
        EventLogger.log_event("deployments_cloud_token_classification")

    click.secho("[+] Model deployed\n", fg="green", bold=True)


@click.group(name="instances")
def instances():
    pass


@click.command(name="ls")
def ls_instances():
    click.secho("\n[+] Collecting all instances\n", fg="blue", bold=True)
    columns, values = Instances.get_instance_types(fmt="table")
    print_table(columns, values)
    EventLogger.log_event("deployments_ls_instances")


deploy.add_command(sequence_classification)
deploy.add_command(question_answering)
deploy.add_command(token_classification)
deploy.add_command(summarization)
deploy.add_command(translation)

deployments.add_command(ls_deployments)
deployments.add_command(launch_deployment)
deployments.add_command(delete_deployment)

instances.add_command(ls_instances)



import click
from stochasticx.client.models import models
from stochasticx.client.datasets import datasets
from stochasticx.client.jobs import jobs
from stochasticx.client.auth import login, me, logout
from stochasticx.client.deployments import instances, deployments
from stochasticx.client.inferences import inference
from stochasticx.client.local import local

from stochasticx.client.stable_diffusion import stable_diffusion
from stochasticx.client.finetuning import finetuning_jobs
from stochasticx.client.preferences import config_command
from stochasticx.client.benchmark import benchmarking
from stochasticx.client.conversion import conversion
import stochasticx


@click.group()
@click.version_option(stochasticx.__version__)
@click.pass_context
def cli(ctx):
    pass


cli.add_command(models)
cli.add_command(datasets)
cli.add_command(jobs)
cli.add_command(login)
cli.add_command(logout)
cli.add_command(me)
cli.add_command(instances)
cli.add_command(deployments)
cli.add_command(inference)
cli.add_command(stable_diffusion)
cli.add_command(finetuning_jobs)
cli.add_command(local)
cli.add_command(config_command)
cli.add_command(benchmarking)
cli.add_command(conversion)



import click
from stochasticx.models.models import Models
from stochasticx.datasets.datasets import Datasets
from stochasticx.jobs.jobs import (
    Job,
    Jobs,
    OptimizationCriteria,
    SequenceClassificationTask,
    QuestionAnsweringTask,
    SummarizationTask,
    TranslationTask,
    TokenClassificationTask,
)
from stochasticx.utils.parse_utils import print_table
from stochasticx.utils.auth_utils import AuthUtils
import sys

from stochasticx.utils.stat_controller import EventLogger


@click.group(name="jobs")
def jobs():
    try:
        AuthUtils.get_auth_headers()
    except:
        sys.exit()
    EventLogger.log_event("jobs")


@click.command(name="ls")
def ls_jobs():
    click.secho("\n[+] Collecting all jobs\n", fg="blue", bold=True)
    columns, values = Jobs.get_jobs(fmt="table")
    print_table(columns, values)
    EventLogger.log_event("jobs_ls")


@click.command(name="inspect")
@click.option("--id", required=True, help="Job ID")
def job_inspect(id):
    job = Jobs.get_job(id)
    click.echo(job)
    EventLogger.log_event("jobs_inspect")


@click.command(name="status")
@click.option("--id", required=True, help="Job ID")
def job_status(id):
    job = Jobs.get_job(id)
    click.echo(job.get_status())
    EventLogger.log_event("jobs_status")


@click.group(name="launch")
def job_launch():
    pass


@click.command(name="sequence_classification")
@click.option("--name", required=True, help="Job name")
@click.option("--model_id", required=True, help="Model ID or S3 path")
@click.option("--dataset_id", required=True, help="Dataset ID or S3 path")
@click.option("--aws_access_key_id", required=False, help="aws_access_key_id if you specify a S3 path for the model or dataset")
@click.option("--aws_secret_access_key", required=False, help="aws_secret_access_key if you specify a S3 path for the model or dataset")
@click.option(
    "--optimization_criteria",
    required=True,
    help="Optimization criteria. It should be latency or lossless",
)
@click.option("--sentence1_column", required=True, help="Sentence 1 column")
@click.option("--sentence2_column", required=True, help="Sentence 2 column")
@click.option("--labels_column", required=True, help="Labels column")
@click.option(
    "--max_seq_length", show_default=True, default=128, help="Maximum sequence length"
)
def job_launch_sequence_classification(
    name,
    model_id,
    dataset_id,
    aws_access_key_id,
    aws_secret_access_key,
    optimization_criteria,
    sentence1_column,
    sentence2_column,
    labels_column,
    max_seq_length,
):
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        model_id = Models.get_model(model_id).get_id()
    
    assert model_id is not None, "Model ID does not exist"
    
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        dataset_id = Datasets.get_dataset(dataset_id).get_id()
        
    assert dataset_id is not None, "Dataset ID does not exist"
    
    assert optimization_criteria in [
        "latency",
        "lossless",
    ], "optimization_criteria should take one of these values: latency or lossless"

    task_type = SequenceClassificationTask(
        sentence1_column=sentence1_column,
        sentence2_column=sentence2_column,
        labels_column=labels_column,
        max_seq_length=max_seq_length,
    )

    job = Job(name=name)
    job.launch_auto(
        model_id=model_id,
        dataset_id=dataset_id,
        task_type=task_type,
        optimization_criteria=OptimizationCriteria.LATENCY,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    EventLogger.log_event("jobs_sequence_classification")


@click.command(name="question_answering")
@click.option("--name", required=True, help="Job name")
@click.option("--model_id", required=True, help="Model ID or S3 path")
@click.option("--dataset_id", required=True, help="Dataset ID or S3 path")
@click.option("--aws_access_key_id", required=False, help="aws_access_key_id if you specify a S3 path for the model or dataset")
@click.option("--aws_secret_access_key", required=False, help="aws_secret_access_key if you specify a S3 path for the model or dataset")
@click.option(
    "--optimization_criteria",
    required=True,
    help="Optimization criteria. It should be latency or lossless",
)
@click.option("--question_column", required=True, help="Question column")
@click.option("--answer_column", required=True, help="Answer column")
@click.option("--context_column", required=True, help="Context column")
@click.option(
    "--max_seq_length", default=128, show_default=True, help="Maximum sequence length"
)
@click.option("--stride", default=30, show_default=True, help="Stride")
def job_launch_question_answering(
    name,
    model_id,
    dataset_id,
    aws_access_key_id,
    aws_secret_access_key,
    optimization_criteria,
    question_column,
    answer_column,
    context_column,
    max_seq_length,
    stride,
):
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        model_id = Models.get_model(model_id).get_id()
    
    assert model_id is not None, "Model ID does not exist"
    
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        dataset_id = Datasets.get_dataset(dataset_id).get_id()
        
    assert dataset_id is not None, "Dataset ID does not exist"
    
    assert optimization_criteria in [
        "latency",
        "lossless",
    ], "optimization_criteria should take one of these values: latency or lossless"

    task_type = QuestionAnsweringTask(
        question_column=question_column,
        answer_column=answer_column,
        context_column=context_column,
        max_seq_length=max_seq_length,
        stride=stride,
    )

    job = Job(name=name)
    job.launch_auto(
        model_id=model_id,
        dataset_id=dataset_id,
        task_type=task_type,
        optimization_criteria=OptimizationCriteria.LATENCY,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    EventLogger.log_event("jobs_question_answering")


@click.command(name="summarization")
@click.option("--name", required=True, help="Job name")
@click.option("--model_id", required=True, help="Model ID or S3 path")
@click.option("--dataset_id", required=True, help="Dataset ID or S3 path")
@click.option("--aws_access_key_id", required=False, help="aws_access_key_id if you specify a S3 path for the model or dataset")
@click.option("--aws_secret_access_key", required=False, help="aws_secret_access_key if you specify a S3 path for the model or dataset")
@click.option(
    "--optimization_criteria",
    required=True,
    help="Optimization criteria. It should be latency or lossless",
)
@click.option("--text_column", required=True, help="Text column")
@click.option("--summary_column", required=True, help="Summary column")
@click.option(
    "--max_source_length", default=256, show_default=True, help="Maximum source length"
)
@click.option(
    "--max_target_length", default=64, show_default=True, help="Maximum target length"
)
@click.option("--lang", default="en", show_default=True, help="Language")
@click.option(
    "--pad_to_max_length",
    default=False,
    show_default=True,
    help="Pad to maximum length",
)
@click.option("--num_beams", default=4, show_default=True, help="Number of beams")
@click.option(
    "--ignore_pad_token_for_loss", default=True, help="Ignore pad token for loss"
)
@click.option("--source_prefix", default="", show_default=True, help="Source prefix")
@click.option(
    "--forced_bos_token", default=None, show_default=True, help="Forced BOS token"
)
def job_launch_summarization(
    name,
    model_id,
    dataset_id,
    aws_access_key_id,
    aws_secret_access_key,
    optimization_criteria,
    text_column,
    summary_column,
    max_source_length,
    max_target_length,
    lang,
    pad_to_max_length,
    num_beams,
    ignore_pad_token_for_loss,
    source_prefix,
    forced_bos_token,
):
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        model_id = Models.get_model(model_id).get_id()
    
    assert model_id is not None, "Model ID does not exist"
    
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        dataset_id = Datasets.get_dataset(dataset_id).get_id()
        
    assert dataset_id is not None, "Dataset ID does not exist"
    
    assert optimization_criteria in [
        "latency",
        "lossless",
    ], "optimization_criteria should take one of these values: latency or lossless"

    task_type = SummarizationTask(
        text_column=text_column,
        summary_column=summary_column,
        max_source_length=max_source_length,
        max_target_length=max_target_length,
        lang=lang,
        pad_to_max_length=pad_to_max_length,
        num_beams=num_beams,
        ignore_pad_token_for_loss=ignore_pad_token_for_loss,
        source_prefix=source_prefix,
        forced_bos_token=forced_bos_token,
    )

    job = Job(name=name)
    job.launch_auto(
        model_id=model_id,
        dataset_id=dataset_id,
        task_type=task_type,
        optimization_criteria=OptimizationCriteria.LATENCY,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    EventLogger.log_event("jobs_question_summarization")


@click.command(name="translation")
@click.option("--name", required=True, help="Job name")
@click.option("--model_id", required=True, help="Model ID or S3 path")
@click.option("--dataset_id", required=True, help="Dataset ID or S3 path")
@click.option("--aws_access_key_id", required=False, help="aws_access_key_id if you specify a S3 path for the model or dataset")
@click.option("--aws_secret_access_key", required=False, help="aws_secret_access_key if you specify a S3 path for the model or dataset")
@click.option(
    "--optimization_criteria",
    required=True,
    help="Optimization criteria. It should be latency or lossless",
)
@click.option("--translation_column", required=True, help="Translation column")
@click.option("--src_lang", default="en", show_default=True, help="Source language")
@click.option("--tgt_lang", default="es", show_default=True, help="Target language")
@click.option(
    "--max_source_length", default=256, show_default=True, help="Maximum source length"
)
@click.option(
    "--max_target_length",
    default=256,
    show_default=True,
    help="Maximum target length length",
)
@click.option(
    "--pad_to_max_length",
    default=False,
    show_default=True,
    help="Pad to maximum length",
)
@click.option("--num_beams", default=4, show_default=True, help="Number of beams")
@click.option(
    "--ignore_pad_token_for_loss",
    default=True,
    show_default=True,
    help="Ignore pad token for loss",
)
@click.option("--source_prefix", default="", show_default=True, help="Source prefix")
@click.option(
    "--forced_bos_token", default=None, show_default=True, help="Forced bos token"
)
def job_launch_translation(
    name,
    model_id,
    dataset_id,
    aws_access_key_id,
    aws_secret_access_key,
    optimization_criteria,
    translation_column,
    src_lang,
    tgt_lang,
    max_source_length,
    max_target_length,
    pad_to_max_length,
    num_beams,
    ignore_pad_token_for_loss,
    source_prefix,
    forced_bos_token,
):
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        model_id = Models.get_model(model_id).get_id()
    
    assert model_id is not None, "Model ID does not exist"
    
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        dataset_id = Datasets.get_dataset(dataset_id).get_id()
        
    assert dataset_id is not None, "Dataset ID does not exist"
    
    assert optimization_criteria in [
        "latency",
        "lossless",
    ], "optimization_criteria should take one of these values: latency or lossless"

    task_type = TranslationTask(
        translation_column=translation_column,
        src_lang=src_lang,
        tgt_lang=tgt_lang,
        max_source_length=max_source_length,
        max_target_length=max_target_length,
        pad_to_max_length=pad_to_max_length,
        num_beams=num_beams,
        ignore_pad_token_for_loss=ignore_pad_token_for_loss,
        source_prefix=source_prefix,
        forced_bos_token=forced_bos_token,
    )

    job = Job(name=name)
    job.launch_auto(
        model_id=model_id,
        dataset_id=dataset_id,
        task_type=task_type,
        optimization_criteria=OptimizationCriteria.LATENCY,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    EventLogger.log_event("jobs_translation")


@click.command(name="token_classification")
@click.option("--name", required=True, help="Job name")
@click.option("--model_id", required=True, help="Model ID or S3 path")
@click.option("--dataset_id", required=True, help="Dataset ID or S3 path")
@click.option("--aws_access_key_id", required=False, help="aws_access_key_id if you specify a S3 path for the model or dataset")
@click.option("--aws_secret_access_key", required=False, help="aws_secret_access_key if you specify a S3 path for the model or dataset")
@click.option(
    "--optimization_criteria",
    required=True,
    help="Optimization criteria. It should be latency or lossless",
)
@click.option("--tokens_column", required=True, help="Tokens column")
@click.option("--domains_column", required=True, help="Domains 2 column")
@click.option("--ner_tags_column", required=True, help="Ner column")
@click.option(
    "--max_seq_length", default=128, show_default=True, help="Maximum sequence length"
)
@click.option(
    "--label_all_tokens", default=False, show_default=True, help="Label all columns"
)
def job_launch_token_classification(
    name,
    model_id,
    dataset_id,
    aws_access_key_id,
    aws_secret_access_key,
    optimization_criteria,
    tokens_column,
    domains_column,
    ner_tags_column,
    max_seq_length,
    label_all_tokens,
):
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        model_id = Models.get_model(model_id).get_id()
    
    assert model_id is not None, "Model ID does not exist"
    
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        dataset_id = Datasets.get_dataset(dataset_id).get_id()
        
    assert dataset_id is not None, "Dataset ID does not exist"
    
    assert optimization_criteria in [
        "latency",
        "lossless",
    ], "optimization_criteria should take one of these values: latency or lossless"

    task_type = TokenClassificationTask(
        tokens_column=tokens_column,
        domains_column=domains_column,
        ner_tags_column=ner_tags_column,
        max_seq_length=max_seq_length,
        label_all_tokens=label_all_tokens,
    )

    job = Job(name=name)
    job.launch_auto(
        model_id=model_id,
        dataset_id=dataset_id,
        task_type=task_type,
        optimization_criteria=OptimizationCriteria.LATENCY,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    EventLogger.log_event("jobs_token_classification")


job_launch.add_command(job_launch_sequence_classification)
job_launch.add_command(job_launch_question_answering)
job_launch.add_command(job_launch_summarization)
job_launch.add_command(job_launch_translation)
job_launch.add_command(job_launch_token_classification)

jobs.add_command(ls_jobs)
jobs.add_command(job_inspect)
jobs.add_command(job_status)
jobs.add_command(job_launch)



from stochasticx.utils.auth_utils import AuthUtils
from stochasticx.constants.urls import get_cloud_url, CloudRoutes
from stochasticx.models.models import Model, Models
from stochasticx.datasets.datasets import Dataset
from stochasticx.utils.parse_utils import parse_date

from typing import List, Dict

import requests


class OptimizationCriteria:
    LATENCY = "latency"
    LOSSLESS = "lossless"


class TaskType:
    def __init__(self):
        pass

    def get_dataset_columns(self):
        raise NotImplementedError(
            "This method should be implemented by the child class"
        )

    def get_task_arguments(self):
        raise NotImplementedError(
            "This method should be implemented by the child class"
        )

    def get_task_type(self):
        raise NotImplementedError(
            "This method should be implemented by the child class"
        )


class SequenceClassificationTask(TaskType):
    def __init__(
        self, sentence1_column, sentence2_column, labels_column, max_seq_length=128
    ):
        self.sentence1_column = sentence1_column
        self.sentence2_column = sentence2_column
        self.labels_column = labels_column
        self.max_seq_length = max_seq_length

    def get_dataset_columns(self):
        return {
            "sentence1_column": self.sentence1_column,
            "sentence2_column": self.sentence2_column,
            "labels_column": self.labels_column,
        }

    def get_task_arguments(self):
        return {"max_seq_length": self.max_seq_length}

    def get_task_type(self):
        return "sequence_classification"


class QuestionAnsweringTask(TaskType):
    def __init__(
        self,
        question_column,
        answer_column,
        context_column,
        max_seq_length=128,
        stride=30,
    ):
        self.question_column = question_column
        self.answer_column = answer_column
        self.context_column = context_column
        self.max_seq_length = max_seq_length
        self.stride = stride

    def get_dataset_columns(self):
        return {
            "question_column": self.question_column,
            "answer_column": self.answer_column,
            "context_column": self.context_column,
        }

    def get_task_arguments(self):
        return {"max_seq_length": self.max_seq_length, "stride": self.stride}

    def get_task_type(self):
        return "question_answering"


class SummarizationTask(TaskType):
    def __init__(
        self,
        text_column,
        summary_column,
        max_source_length=256,
        max_target_length=64,
        lang="en",
        pad_to_max_length=False,
        num_beams=4,
        ignore_pad_token_for_loss=True,
        source_prefix="",
        forced_bos_token=None,
    ):
        self.text_column = text_column
        self.summary_column = summary_column
        self.max_source_length = max_source_length
        self.max_target_length = max_target_length
        self.lang = lang
        self.pad_to_max_length = pad_to_max_length
        self.num_beams = num_beams
        self.ignore_pad_token_for_loss = ignore_pad_token_for_loss
        self.source_prefix = source_prefix
        self.forced_bos_token = forced_bos_token

    def get_dataset_columns(self):
        return {"text_column": self.text_column, "summary_column": self.summary_column}

    def get_task_arguments(self):
        return {
            "max_source_length": self.max_source_length,
            "max_target_length": self.max_target_length,
            "lang": self.lang,
            "pad_to_max_length": self.pad_to_max_length,
            "num_beams": self.num_beams,
            "ignore_pad_token_for_loss": self.ignore_pad_token_for_loss,
            "source_prefix": self.source_prefix,
            "forced_bos_token": self.forced_bos_token,
        }

    def get_task_type(self):
        return "summarization"


class TranslationTask(TaskType):
    def __init__(
        self,
        translation_column,
        src_lang="en",
        tgt_lang="es",
        max_source_length=256,
        max_target_length=256,
        pad_to_max_length=False,
        num_beams=4,
        ignore_pad_token_for_loss=True,
        source_prefix="",
        forced_bos_token=None,
    ):
        self.translation_column = translation_column
        self.src_lang = src_lang
        self.tgt_lang = tgt_lang
        self.max_source_length = max_source_length
        self.max_target_length = max_target_length
        self.pad_to_max_length = pad_to_max_length
        self.num_beams = num_beams
        self.ignore_pad_token_for_loss = ignore_pad_token_for_loss
        self.source_prefix = source_prefix
        self.forced_bos_token = forced_bos_token

    def get_dataset_columns(self):
        return {"translation_column": self.translation_column}

    def get_task_arguments(self):
        return {
            "src_lang": self.src_lang,
            "tgt_lang": self.tgt_lang,
            "max_source_length": self.max_source_length,
            "max_target_length": self.max_target_length,
            "pad_to_max_length": self.pad_to_max_length,
            "num_beams": self.num_beams,
            "ignore_pad_token_for_loss": self.ignore_pad_token_for_loss,
            "source_prefix": self.source_prefix,
            "forced_bos_token": self.forced_bos_token,
        }

    def get_task_type(self):
        return "translation"


class TokenClassificationTask(TaskType):
    def __init__(
        self,
        tokens_column,
        domains_column,
        ner_tags_column,
        max_seq_length=128,
        label_all_tokens=False,
    ):
        self.tokens_column = tokens_column
        self.domains_column = domains_column
        self.ner_tags_column = ner_tags_column
        self.max_seq_length = max_seq_length
        self.label_all_tokens = label_all_tokens

    def get_dataset_columns(self):
        return {
            "tokens_column": self.tokens_column,
            "domains_column": self.domains_column,
            "ner_tags_column": self.ner_tags_column,
        }

    def get_task_arguments(self):
        return {
            "max_seq_length": self.max_seq_length,
            "label_all_tokens": self.label_all_tokens,
        }

    def get_task_type(self):
        return "token_classification"


class Job:
    def __init__(self, name):
        self.id = None
        self.name = name
        self.original_model = None
        self.optimization_type = None
        self.optimization_criteria = None
        self.optimized_models = []
        self.job_status = None
        self.created_at = None

    def _populate_job(self, info_from_api):
        self.id = info_from_api.get("id")
        self.name = info_from_api.get("name")

        original_model_id = None
        original_model_data = info_from_api.get("model")
        if original_model_data is not None and isinstance(original_model_data, dict):
            original_model_id = original_model_data.get("id")
        self.original_model = Models.get_model(original_model_id)

        self.optimization_type = info_from_api.get("optimizationType")
        self.optimization_criteria = info_from_api.get("optimizationCriteria")
        self.job_status = info_from_api.get("jobStatus")
        self.created_at = info_from_api.get("createdAt")

        if info_from_api.get("models") is not None and isinstance(
            info_from_api.get("models"), list
        ):
            for optimized_model_id in info_from_api.get("models"):
                optimized_model = Models.get_optimized_model(optimized_model_id)
                self.optimized_models.append(optimized_model)

    def sync(self):
        if self.id is not None:
            temp_job = Jobs.get_job(self.id)
            self.name = temp_job.name
            self.original_model = temp_job.original_model
            self.optimization_type = temp_job.optimization_type
            self.optimization_criteria = temp_job.optimization_criteria
            self.optimized_models = temp_job.optimized_models
            self.job_status = temp_job.job_status
            self.created_at = temp_job.created_at

    def launch_auto(
        self,
        model_id: str,
        dataset_id: str,
        task_type: TaskType,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        optimization_criteria: str = OptimizationCriteria.LATENCY,
    ):
        assert isinstance(task_type, TaskType), "The task type is not valid"

        job_parameters = {
            "dataset": model_id,
            "model": dataset_id,
            "modelCloud": "s3",
            "jobName": self.name,
            "optimizationCriteria": optimization_criteria,
            "selectedColumns": task_type.get_dataset_columns(),
            "taskArguments": task_type.get_task_arguments(),
            "taskType": task_type.get_task_type(),
            "optimizationType": "auto",
            "finetuningArguments": {},
            "knowledgeDistillationArguments": {},
            "instanceType": "g4dn.xlarge",
            "isCPU": False,
            "isGPU": True,
        }
        
        if aws_access_key_id is not None and aws_secret_access_key is not None:
            job_parameters["authentication"] = {
                "cloud": "s3",
                "details": {
                    "aws_access_key_id": aws_access_key_id,
                    "aws_secret_access_key": aws_secret_access_key
                }
            }

        auth_header = AuthUtils.get_auth_headers()
        r = requests.post(
            get_cloud_url(CloudRoutes.JOBS_URL),
            headers=auth_header,
            json=job_parameters,
        )
        r.raise_for_status()

        data = r.json().get("data")
        if data is not None:
            self.id = data.get("id")

    def get_status(self):
        self.sync()

        return self.job_status

    def to_table(self):
        columns = [
            "Id",
            "Name",
            "Status",
            "Created at",
            "Optimization type",
            "Optimization criteria",
        ]

        values = [
            self.id,
            self.name,
            self.job_status,
            parse_date(self.created_at),
            self.optimization_type,
            self.optimization_criteria,
        ]

        return columns, values

    def __str__(self):
        return "Job ID: {} ; Name: {} ; Status: {} ; Created at: {} ; Optimization type: {} ; Optimization criteria: {}".format(
            self.id,
            self.name,
            self.job_status,
            parse_date(self.created_at),
            self.optimization_type,
            self.optimization_criteria,
        )


class Jobs:
    @staticmethod
    def get_jobs(fmt=None):
        jobs = []

        auth_header = AuthUtils.get_auth_headers()
        response = requests.get(
            get_cloud_url(CloudRoutes.JOBS_URL), headers=auth_header
        )
        jobs_data = response.json().get("data")

        if jobs_data is not None:
            for job_data in jobs_data:
                job = Job(name=None)

                job._populate_job(job_data)
                jobs.append(job)

            if fmt == "table":
                columns = []
                values = []

                if len(jobs) > 0:
                    columns, _ = jobs[0].to_table()

                for job in jobs:
                    _, vals = job.to_table()
                    values.append(vals)

                return columns, values

        return jobs

    @staticmethod
    def get_job(job_id):
        url = get_cloud_url(CloudRoutes.JOBS_URL.formatting(job_id))

        auth_header = AuthUtils.get_auth_headers()
        response = requests.get(url, headers=auth_header)
        job_data = response.json().get("data")

        if job_data is not None:
            job_data = job_data.get("job")

        if job_data is not None:
            job = Job(name=job_data.get("name"))

            job._populate_job(job_data)

            return job

        return None
